@@ -37,6 +37,7 @@ from .tdms import NoData
 from dastools import __version__
 import dastools.archive as da
 import inspect
+from tqdm import tqdm
 
 
 def nslc(dataheader: dict) -> str:
@@ -104,7 +105,8 @@ def main():
                         help='Verbosity in the output.',
                         choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO',
                                  'DEBUG'],
-                        default='WARNING')
+                        default='INFO')
+    parser.add_argument('--logout', help='Name of the log file', default='output.log', type=str)
     parser.add_argument('-d', '--directory',
                         help='Directory where files are located (default: ".")',
                         default='.')
@@ -144,7 +146,7 @@ def main():
     if args.metadata:
         logging.basicConfig(level=args.loglevel, stream=sys.stdout)
     else:
-        logging.basicConfig(level=args.loglevel)
+        logging.basicConfig(filename=args.logout, encoding='utf-8', level=args.loglevel)
 
     logs = logging.getLogger('OpenFile')
     logs.setLevel(args.loglevel)
@@ -183,10 +185,12 @@ def main():
     with td:
         curstream = None
 
-        for data in td:
+        progress = tqdm(td, total=td.chunks())
+        for data in progress:
             if args.metadata:
                 printmetadata(data)
             else:
+                progress.set_postfix_str('(sta:%s) %s' % (data[1].station, data[1].starttime.isoformat()))
                 logs.debug(data[1])
 
                 if curstream != nslc(data[1]):
