@@ -1,32 +1,55 @@
-###################################################################################################
-# (C) 2021 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences, Potsdam, Germany  #
-#                                                                                                 #
-# This file is part of dastools.                                                                  #
-#                                                                                                 #
-# dastools is free software: you can redistribute it and/or modify it under the terms of the GNU  #
-# General Public License as published by the Free Software Foundation, either version 3 of the    #
-# License, or (at your option) any later version.                                                 #
-#                                                                                                 #
-# dastools is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without   #
-# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   #
-# General Public License for more details.                                                        #
-#                                                                                                 #
-# You should have received a copy of the GNU General Public License along with this program. If   #
-# not, see https://www.gnu.org/licenses/.                                                         #
-###################################################################################################
+"""TDMS module from dastools.
+
+This file is part of dastools.
+
+dastools is free software: you can redistribute it and/or modify it under the terms of the GNU
+General Public License as published by the Free Software Foundation, either version 3 of the
+License, or (at your option) any later version.
+
+dastools is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
+even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License along with this program. If
+not, see https://www.gnu.org/licenses/.
+
+   :Copyright:
+       2021 Helmholtz Centre Potsdam GFZ German Research Centre for Geosciences, Potsdam, Germany
+   :License:
+       GPLv3
+   :Platform:
+       Linux
+
+.. moduleauthor:: Javier Quinteros <javier@gfz-potsdam.de>, GEOFON, GFZ Potsdam
+"""
 
 import logging
 import datetime
 import os
 import struct
 from obspy import UTCDateTime
+from obspy.core.trace import Stats
+from obspy.core.util.attribdict import AttribDict
 import numpy as np
 from math import floor
 from math import ceil
 from numbers import Number
 
 
-def tup2time(fraction, seconds):
+def tup2time(fraction: int, seconds: int) -> datetime.datetime:
+    """Convert a tuple of fraction and seconds in a timestamp
+
+    :param fraction: The least significant 64 bits should be interpreted as a 64-bit unsigned integer.
+                     It represents the number of 2-64 seconds after the whole seconds specified in the most significant 64-bits.
+    :type fraction: int
+    :param seconds: The most significant 64 bits should be interpreted as a 64-bit signed two's complement integer.
+                    It represents the number of whole seconds after the Epoch 01/01/1904 00:00:00.00 UTC.
+    :type seconds: int
+    :returns: Datetime of the timestamp
+    :rtype: datetime.datetime
+
+    .. todo:: Check in case of little endian if the parameters are not swapped
+    """
     # logs = logging.getLogger('tup2time')
     # logs.debug('seconds: %s' % seconds)
     # logs.debug('fraction: %s' % fraction)
@@ -39,15 +62,25 @@ def tup2time(fraction, seconds):
 
 
 class PotentialGap(Exception):
+    """Exception to signal that a gap has been found"""
     pass
 
 
 class NoData(Exception):
+    """Exception to signal that there is no data"""
     pass
 
 
 class TDMS(object):
-    """Class to read and export seismic waveforms in TDMS format"""
+    """Class to read, process and export seismic waveforms in TDMS format
+
+.. note::
+    Some examples with PDN_1km show that for 16 files of 44 MB (704 MB) we see the following
+    With simple encoding of I16 and a record of 4096 bytes we use 720MB
+    With STEIM2, a data type of I32 and a record of 4096 bytes we use 504 MB
+    With STEIM2, a data type of I32, a decimation factor of 5, and a record of 4096 bytes we use 109 MB
+
+    """
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         """Method which close open resources after using the syntax 'with object:' and use it inside"""
@@ -55,20 +88,21 @@ class TDMS(object):
             self.__fi.close()
 
     def __enter__(self):
-        """Method which allows to use the syntax 'with object:' and use it inside"""
+        """Method which allows to use the syntax 'with object:' and use it inside
 
-        # Create a buffer space to store the signal coefficients to be
-        # convoluted during the decimation
-        # for channel in range(self.__chstart, self.__chstop + 1, self.__chstep):
+        Create a buffer space to store the signal coefficients to be convoluted during the decimation
+        """
         for channel in self.__channels:
             logging.debug('Create empty buffer for channel %s' % channel)
             self.__buffer[channel] = None
 
         return self
 
-    def __init__(self, filename, directory='.', chstart=0, chstop=None, chstep=1, channels=None,
-                 starttime=None, endtime=None, iterate='D', decimate=1, firfilter='fir235',
-                 loglevel='INFO'):
+    def __init__(self, filename: str, directory: str = '.', chstart: int = 0, chstop: int = None,
+                 chstep: int = 1, channels: list = None, starttime: datetime.datetime = None,
+                 endtime: datetime.datetime = None, iterate: str = 'D', decimate: int = 1,
+                 firfilter: str = 'fir235', networkcode: str = 'XX', channelcode: str = 'FSF',
+                 loglevel: str = 'INFO'):
         """Initialize the TDMS object selecting the data, channels and decimation
 
         :param filename: Experiment to read and process. Usually the first part of the filenames
@@ -93,8 +127,16 @@ class TDMS(object):
         :type decimate: int
         :param firfilter: Filter to apply in case of decimation (fir235 is the only option)
         :type firfilter: str
+        :param networkcode: Network code of the experiment. It has to be two alphanumeric characters
+        :type networkcode: str
+        :param channelcode: Channel code of the experiment. It has to be three alphanumeric characters
+        :type channelcode: str
         :param loglevel: Verbosity in the output
         :type loglevel: str
+        :raise TypeError: If chstart, or chstop, or chstep are not int. If channels is not a list, or networkcode is
+        not a 2 characters code, or channelcode is not a 3 characters code.
+        :raise Exception: If channels is empty.
+        :raise NoData: If there is no more data available
         """
 
         # Log level
@@ -109,20 +151,30 @@ class TDMS(object):
 
         # Selection of channels
         if not isinstance(chstart, Number) or not isinstance(chstep, Number):
-            logs.error('chstart and chstep must be numbers and preferably ints')
-            raise TypeError('chstart and chstep must be numbers and preferably ints')
+            logs.error('chstart and chstep must be ints')
+            raise TypeError('chstart and chstep must be ints')
 
         if not isinstance(chstop, Number) and chstop is not None:
-            logs.error('chstop must be a number or None')
-            raise TypeError('chstop must be a number or None')
+            logs.error('chstop must be an int or None')
+            raise TypeError('chstop must be an int or None')
 
         if not isinstance(channels, list) and channels is not None:
             logs.error('channels must be a list of numbers or None')
             raise TypeError('channels must be a list of numbers or None')
 
+        if not isinstance(networkcode, str) or len(networkcode) != 2:
+            logs.error('Network code has to be two alphanumeric characters')
+            raise TypeError('Network code has to be two alphanumeric characters')
+
+        if not isinstance(channelcode, str) or len(channelcode) != 3:
+            logs.error('Channel code has to be three alphanumeric characters')
+            raise TypeError('Channel code has to be three alphanumeric characters')
+
         self.__chstart = chstart
         self.__chstop = chstop
         self.__chstep = chstep
+        self.__networkcode = networkcode
+        self.__channelcode = channelcode
 
         # channels has priority. Otherwise, chstart, chstop and chstep are used
         if channels is not None and isinstance(channels, list):
@@ -162,11 +214,20 @@ class TDMS(object):
             if not file.startswith(filename):
                 continue
 
-            dt = datetime.datetime.strptime(file[len(filename):-len('.tdms')], '_%Z_%Y%m%d_%H%M%S.%f')
+            # Check that the format of the filename is exactly as expected. Otherwise, send a warning
+            try:
+                dt = datetime.datetime.strptime(file[len(filename):-len('.tdms')], '_%Z_%Y%m%d_%H%M%S.%f')
+            except ValueError:
+                logs.warning('Unexpected format after experiment name! Skipping... (%s)' % file[len(filename):])
+                continue
             self.__available.append({'dt': dt, 'name': file, 'samples': None})
             if self.__twstart is None:
                 self.__twstart = dt
 
+        if not len(self.__available):
+            logs.warning('No files found with proper filename.')
+            raise NoData()
+
         # Set the time window selection to the minimum datetime found
         if self.__twstart < self.__available[0]['dt']:
             self.__twstart = self.__available[0]['dt']
@@ -183,9 +244,6 @@ class TDMS(object):
         # Keys will be channel number and the values will be np.arrays
         self.__buffer = dict()
 
-        # Datatype of each channel
-        self.__datatypes = dict()
-
         # Dictionary to save the metadata defined in the file
         self.metadata = dict()
 
@@ -230,6 +288,7 @@ class TDMS(object):
         self.__endian = None
         self.__datatype = None
         self.__datatypesize = None
+        self.__outdatatype = None
         self.numchannels = None
         self.__samples = None
         self.__samplestart = None
@@ -242,6 +301,12 @@ class TDMS(object):
             raise NoData()
 
     def __select_file(self):
+        """Select a file from the experiment based on the status of the object
+
+        :raise Exception: If data not available in the specified time window. If the header does not
+        indicate that the file is a TDMS format
+        :raise IndexError: If the last file has already been processed or the start is greater than end
+        """
         logs = logging.getLogger('Select file')
         logs.setLevel(self.__loglevel)
 
@@ -316,6 +381,7 @@ class TDMS(object):
 
         if self.__segmentOffset == self.__FF64b:
             logs.error('Severe problem while writing data (crash, power outage)')
+            raise Exception('Corrupted file due to crash or power outage')
 
         if hasmetadata and not self.__dataOffset:
             logs.error('Flag indicates Metadata but its length is 0!')
@@ -338,30 +404,51 @@ class TDMS(object):
 
     def __search_data(self):
         """
-        :raise: IndexError
+        Select a file to work with, read its metadata and calculate samples to read
+
+        :raise IndexError: If the last file has already been processed or the start is greater than end
         """
 
         while True:
             # Loop through files until there is nothing else (IndexError)
-            self.__select_file()
+            try:
+                self.__select_file()
+            except IndexError:
+                raise
+            except Exception as e:
+                logging.warning('Skipping file because of %s' % str(e))
+                self.__currentfile += 1
+                continue
 
             # Read the metadata and calculate samples to read
             # Skip to the next file if there is a gap
             try:
                 self.__readmetadata()
                 return
+            except NoData:
+                logging.error('No Data was found in the metadata definition')
+                self.__currentfile += 1
             except PotentialGap:
-                logging.debug('Potential gap detected!')
+                logging.warning('Potential gap detected!')
                 self.__currentfile += 1
 
     # Rename to make it accessible to users
     def reset(self):
+        """Reset the status of the object and start the read again
+
+        :raise IndexError: If the last file has already been processed or the start is greater than end
+        """
         self.__twstart = self.__origstarttime
         self.__twend = self.__origendtime
         self.__currentfile = None
         self.__search_data()
 
     def __readmetadata(self):
+        """Read metadata of the current file
+
+        :raise NoData: if datatype definition is not found in any channel, or if the data type is not
+        supported, or if no valid channel IDs could be selected
+        """
         # Metadata
         logs = logging.getLogger('Read Metadata')
         # handler = logging.StreamHandler(sys.stdout)
@@ -393,6 +480,16 @@ class TDMS(object):
                 try:
                     if self.sampling_rate is None:
                         self.sampling_rate = self.metadata[obj]['SamplingFrequency[Hz]']
+                        # Check for decimation and take into account the filter length
+                        if self.__decimate != 1:
+                            tap = (len(self.__filter) - 1) / (2.0 * self.sampling_rate)
+                            # If there is decimation adjust the start and end times to include the tapering
+                            # and later keep exactly what user requests
+                            self.__twstart -= datetime.timedelta(seconds=tap)
+                            if self.__twend is not None:
+                                self.__twend += datetime.timedelta(seconds=tap)
+                            logs.debug('Readjust start and end time to accommodate the filter length: %s - %s' %
+                                       (self.__twstart, self.__twend))
                 except KeyError:
                     pass
 
@@ -431,10 +528,17 @@ class TDMS(object):
 
         # Set the data type as numpy expects it
         if datatype is None:
-            raise Exception('datatype definition not found in any channel!')
+            raise NoData('datatype definition not found in any channel!')
 
-        if self.__data2mask[datatype][0] == 'h':
+        if self.__data2mask[datatype][0] == 'b':
+            self.__datatype = '%ci1' % self.__endian
+            self.__outdatatype = '%ci4' % self.__endian
+        elif self.__data2mask[datatype][0] == 'h':
             self.__datatype = '%ci2' % self.__endian
+            self.__outdatatype = '%ci4' % self.__endian
+        elif self.__data2mask[datatype][0] == 'i':
+            self.__datatype = '%ci4' % self.__endian
+            self.__outdatatype = '%ci4' % self.__endian
         elif self.__data2mask[datatype][0] == 'f':
             self.__datatype = '%cf4' % self.__endian
         else:
@@ -496,7 +600,11 @@ class TDMS(object):
         # newobjects = struct.unpack('%cI' % self.__endian, self.__fi.read(4))[0]
 
     def __iter__(self):
-        """Iterate through data (or metadata) and filter and decimate if requested"""
+        """Iterate through data (or metadata) and filter and decimate if requested
+
+        :returns: Data and attributes for the header, or metadata
+        :rtype: tuple(numpy.array, obspy.core.trace.Stats) or dict
+        """
 
         # Create logger
         logs = logging.getLogger('__iter__')
@@ -509,8 +617,8 @@ class TDMS(object):
             # In the case that we iterate to get data
             # If no decimation is needed
             if self.__decimate == 1:
-                for info in self.__iter_data__():
-                    yield info
+                for data, stats in self.__iter_data__():
+                    yield data.astype(self.__outdatatype), stats
             else:
                 # Use an input buffer to store the data coming in chunks from __iter_data__
                 inbuf = dict()
@@ -523,7 +631,8 @@ class TDMS(object):
                 # Time of the first sample in the next file
                 expectedtime = dict()
                 # Flag to know if we are at the beginning of the chunk
-                startofchunk = True
+                # There should be one entry per channel with a default value of True
+                startofchunk = dict()
 
                 for data, stats in self.__iter_data__():
                     ch = int(stats['station'])
@@ -539,16 +648,21 @@ class TDMS(object):
 
                             # Flush outbuf?
                             if ch in outbuf:
-                                leftover = len(nodecimation[ch]) % 5
-                                # outbuf[ch]['stats']['starttime'] += 1/outbuf[ch]['stats']['sampling_rate']
-                                outbuf[ch]['stats']['npts'] = 1
-                                yield nodecimation[ch][-leftover::5], outbuf[ch]['stats']
-                                logs.debug('Flushing: %s %s' % (nodecimation[ch][-leftover::5], outbuf[ch]['stats']))
+                                if 'data' in outbuf[ch]:
+                                    # Set number of points
+                                    outbuf[ch]['stats']['npts'] = len(outbuf[ch]['data'])
+                                    # outbuf[ch]['stats']['starttime'] += 1/outbuf[ch]['stats']['sampling_rate']
+                                    # outbuf[ch]['stats']['npts'] = 1
+                                    logs.debug('Flushing: %s %s' % (outbuf[ch]['stats'], outbuf[ch]['data']))
+                                    yield outbuf[ch]['data'], outbuf[ch]['stats']
+                                # Remove all data and stats from the output buffer after a GAP
                                 del outbuf[ch]
+                            else:
+                                logs.debug('Nothing to flush after GAP')
                             # self.__buffer[ch] = None
                             del expectedtime[ch]
                             # Start the processing of a new chunk after the gap
-                            startofchunk = True
+                            startofchunk[ch] = True
 
                     expectedtime[ch] = stats['starttime'] + stats['npts']/self.sampling_rate
 
@@ -561,18 +675,19 @@ class TDMS(object):
                     else:
                         inbuf[ch]['data'] = np.append(inbuf[ch]['data'], data)
 
-                    # Store values resulting from the convolution
+                    # Set values of stats for the output buffer resulting from the convolution
                     if ch not in outbuf:
                         # Initialize an idx and array with the first chunk of data
                         outbuf[ch] = {'stats': inbuf[ch]['stats'].copy()}
 
                         # Modify the starting time only if it is the beginning of the signal
-                        if startofchunk:
+                        if (ch not in startofchunk) or startofchunk[ch]:
                             outbuf[ch]['stats']['starttime'] += (len(self.__filter)-1)/(2 * outbuf[ch]['stats']['sampling_rate'])
-                            startofchunk = False
+                            startofchunk[ch] = False
 
                         # Change the headers to reflect the decimation
-                        outbuf[ch]['stats']['sampling_rate'] = stats['sampling_rate']/5
+                        # Reduce the sampling rate by the decimation factor
+                        outbuf[ch]['stats']['sampling_rate'] = stats['sampling_rate']/self.__decimate
                         # print('outbuf new: %s' % outbuf[ch]['stats'])
 
                     # If there are not enough components move to the next chunk
@@ -581,24 +696,35 @@ class TDMS(object):
 
                     # Convolution of inbuf with filter and then leave the last 234 values (len(filter)-1)
                     # This convolution is performed on a long array (inbuf) and a default filter with 235 components
-                    nodecimation[ch] = np.convolve(inbuf[ch]['data'], self.__filter, 'valid').astype(data.dtype)
+                    # We save the result of the convolution as int32 to be able to use STEIM2 later
+                    nodecimation[ch] = np.convolve(inbuf[ch]['data'], self.__filter, 'valid').astype(self.__outdatatype)
                     logs.debug('filtered: %d components' % len(nodecimation[ch]))
                     logs.debug('filtered[%d][:11] %s' % (ch, nodecimation[ch][:11]))
                     logs.debug('filtered[%d][-11:] %s' % (ch, nodecimation[ch][-11:]))
 
                     # Check if we can copy as many components as a multiple of the decimation factor
-                    leftover = len(nodecimation[ch]) % 5
-                    logs.debug('filtered: leave %d components for next iteration %s' % (leftover, nodecimation[ch][-leftover:]))
+                    leftover = len(nodecimation[ch]) % self.__decimate
+                    logs.debug('filtered: leave %d components for next iteration %s' %
+                               (leftover, nodecimation[ch][-leftover:]))
+
                     if leftover:
                         if 'data' not in outbuf[ch]:
-                            outbuf[ch]['data'] = nodecimation[ch][:-leftover][::5]
+                            # Take samples each "self.__decimate" components
+                            outbuf[ch]['data'] = nodecimation[ch][:-leftover:self.__decimate]
                         else:
-                            outbuf[ch]['data'] = np.append(outbuf[ch]['data'], nodecimation[ch][:-leftover][::5])
+                            # Add samples each "self.__decimate" components
+                            outbuf[ch]['data'] = np.append(outbuf[ch]['data'],
+                                                           nodecimation[ch][:-leftover:self.__decimate])
                     else:
                         if 'data' not in outbuf[ch]:
-                            outbuf[ch]['data'] = nodecimation[ch][::5]
+                            # Take samples each "self.__decimate" components
+                            outbuf[ch]['data'] = nodecimation[ch][::self.__decimate]
                         else:
-                            outbuf[ch]['data'] = np.append(outbuf[ch]['data'], nodecimation[ch][::5])
+                            # Add samples each "self.__decimate" components
+                            outbuf[ch]['data'] = np.append(outbuf[ch]['data'], nodecimation[ch][::self.__decimate])
+
+                    # Remove filtered signal after copying it to the output buffer
+                    del nodecimation[ch]
 
                     logs.debug('outbuf[%d][:11] %s' % (ch, outbuf[ch]['data'][:11]))
                     logs.debug('outbuf[%d][-11:] %s' % (ch, outbuf[ch]['data'][-11:]))
@@ -615,7 +741,7 @@ class TDMS(object):
                     # If there is enough data
                     if len(outbuf[ch]['data']) > 2000:
                         outbuf[ch]['stats']['npts'] = len(outbuf[ch]['data'])
-                        # print('outbuf: %s npts' % outbuf[ch]['index'])
+                        logs.debug('Sending: %s %s' % (outbuf[ch]['stats'], outbuf[ch]['data']))
                         yield outbuf[ch]['data'], outbuf[ch]['stats']
                         # Reset outbuf with an empty array and the next starttime (in headers)
                         outbuf[ch]['stats']['starttime'] += len(outbuf[ch]['data']) / outbuf[ch]['stats']['sampling_rate']
@@ -625,10 +751,17 @@ class TDMS(object):
                 for ch in outbuf:
                     if ('data' in outbuf[ch]) and len(outbuf[ch]['data']):
                         outbuf[ch]['stats']['npts'] = len(outbuf[ch]['data'])
+                        logs.debug('Sending: %s %s' % (outbuf[ch]['stats'], outbuf[ch]['data']))
                         yield outbuf[ch]['data'], outbuf[ch]['stats']
+                        outbuf[ch]['stats']['starttime'] += len(outbuf[ch]['data']) / outbuf[ch]['stats']['sampling_rate']
+                        del outbuf[ch]['data']
 
     def __iter_data__(self):
-        """Read data from files based on channel selection"""
+        """Read data from files based on channel selection
+
+        :return: Data and attributes for the header
+        :rtype: tuple(numpy.array, obspy.core.trace.Stats)
+        """
         # Data
         logs = logging.getLogger('Iterate Data')
 
@@ -636,14 +769,23 @@ class TDMS(object):
             data = self.__readdata(channels=self.__channels)
             # Loop through channels
             for ch in self.__channels:
-                stats = {'network': 'XX', 'station': '%05d' % ch, 'location': '',
-                         'channel': 'FH1', 'npts': len(data[ch]),
-                         'sampling_rate': self.sampling_rate,
-                         'starttime': UTCDateTime(self.__twstart),
-                         'mseed': {'byteorder': self.__endian,
-                                   'dataquality': 'D',
-                                   'record_length': 512}}
+                stats = Stats()
+                stats.network = self.__networkcode
+                stats.station = '%05d' % ch
+                stats.location = ''
+                stats.channel = self.__channelcode
+                stats.sampling_rate = self.sampling_rate
+                stats.npts = len(data[ch])
+                stats.starttime = UTCDateTime(self.__twstart)
+                stats.mseed = AttribDict()
+                stats.mseed.byteorder = self.__endian
+                stats.mseed.dataquality = 'D'
+                stats.mseed.record_length = 4096
+                stats.mseed.blkt1001 = AttribDict()
+                stats.mseed.blkt1001.timing_quality = 100
+
                 logs.debug('Data length: %d; First component: %s' % (len(data[ch]), data[ch][0]))
+                logs.debug('Stats: %s' % (stats,))
                 yield data[ch], stats
 
             # No more data in this file. Skip to the next one.
@@ -655,6 +797,11 @@ class TDMS(object):
                 break
 
     def __iter_metadata__(self):
+        """Read metadata from files based on channel selection
+
+        :return: Metadata from selected channels
+        :rtype: dict
+        """
         # Metadata
         # logs = logging.getLogger('Iterate Metadata')
 
@@ -674,12 +821,18 @@ class TDMS(object):
             except IndexError:
                 break
 
-    def __readstring(self):
-        """All strings in TDMS files, such as object paths, property names, property values, and raw data values,
+    def __readstring(self) -> str:
+        """Read a string from a TDMS file
+
+        All strings in TDMS files, such as object paths, property names, property values, and raw data values,
         are encoded in UTF-8 Unicode. All of them, except for raw data values, are preceded by a 32-bit unsigned
         integer that contains the length of the string in bytes, not including the length value itself. Strings in
         TDMS files can be null-terminated, but since the length information is stored, the null terminator will be
         ignored when you read from the file.
+
+        :return: String read
+        :rtype: str
+
         """
         # logs = logging.getLogger('readstring')
         strlen = struct.unpack('%cI' % self.__endian, self.__fi.read(4))
@@ -687,6 +840,11 @@ class TDMS(object):
         return self.__fi.read(strlen[0]).decode()
 
     def __readvalue(self):
+        """Read a value from a TDMS file
+
+        :return: Value
+        :rtype: char or int or float or str or boolean or datetime.datetime
+        """
         # logs = logging.getLogger('readvalue')
 
         datatype = self.__readdatatype()
@@ -712,11 +870,18 @@ class TDMS(object):
         # logs.debug('result: %s' % result)
         return result
 
-    def __readproperties(self, result=None):
-        """For each property, the following information is stored:
+    def __readproperties(self, result: dict = None) -> dict:
+        """Read one or many properties and save them in a dictionary
+
+        :param result: Where to store the properties read
+        :type result: dict
+        :return: Properties read
+        :rtype: dict
+
+        For each property, the following information is stored:
             Name (string)
             Data type (tdsDataType)
-            Value (numerics stored binary, strings stored as explained above).
+            Value (numeric stored binary, string stored as explained above).
         """
         logs = logging.getLogger('readproperties')
 
@@ -736,17 +901,19 @@ class TDMS(object):
 
         return result
 
-    def __readdatatype(self):
+    def __readdatatype(self) -> int:
+        """Read data type of a field in TDMS format"""
         return struct.unpack('%cI' % self.__endian, self.__fi.read(4))[0]
 
-    def __readdata(self, channels=None):
+    def __readdata(self, channels: list = None) -> dict:
         """Read a chunk of data from the specified channels.
         Update the attribute __samplecur
 
         :param channels: List of channel numbers to read data from
         :type channels: list
-        :return: Dictionary with channel as key and a numpy array with data as value
+        :return: Dictionary with channel number as key and a numpy array with data as value
         :rtype: dict
+        :raise Exception: if trying to read data from an originally unselected channel
         """
 
         logs = logging.getLogger('Read data')
@@ -790,13 +957,39 @@ class TDMS(object):
         for ch in channels:
             result[ch] = np.zeros((numsamples,), dtype=self.__datatype)
 
+        # Collect potential errors for later logging
+        problematicchannels = set()
+
         for sample in range(numsamples):
             # Read from all channels and select the specific one with an index (channel)
             allchannels = np.fromfile(self.__fi, dtype=self.__datatype, count=self.numchannels)
-            for ch in channels:
-                result[ch][sample] = allchannels[ch]
 
+            for ch in channels:
+                # Check that the channels to read actually exists
+                # Some experiments were found in which after restart there are less channels than at the beginning
+                if ch < len(allchannels):
+                    result[ch][sample] = allchannels[ch]
+                else:
+                    problematicchannels.add(ch)
+
+        if len(problematicchannels):
+            logs.error('Problems with channels: %s' % ','.join(map(str, problematicchannels)))
         # Update the current sample number based on the number of components read
         self.__samplecur += numsamples
 
         return result
+
+    def chunks(self) -> int:
+        """Estimate number of chunks to iterate"""
+        estimatedfiles = 0
+
+        dts = [fi['dt'] for fi in self.__available]
+        for idx in range(len(dts)):
+            if self.__twstart <= dts[idx]:
+                if (self.__twend is None) or (dts[idx] < self.__twend):
+                    estimatedfiles += 1
+                    continue
+            if dts[idx] <= self.__twstart < dts[idx+1]:
+                estimatedfiles += 1
+
+        return len(self.__channels)*estimatedfiles
